/***************************************************************************
**                                                                        **
**  hmFem - Klasse														  **
**	Numerische Loesung von 1D Randwertproblemen (RWP)					  **
**	mit Hilfe der Methode der Finiten Elemente  (FEM)					  **
**  Copyright (C) 2015 Andreas Beinstingel                                **
**                                                                        **
**  This program is free software: you can redistribute it and/or modify  **
**  it under the terms of the GNU General Public License as published by  **
**  the Free Software Foundation, either version 3 of the License, or     **
**  (at your option) any later version.                                   **
**                                                                        **
**  This program is distributed in the hope that it will be useful,       **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of        **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         **
**  GNU General Public License for more details.                          **
**                                                                        **
**  You should have received a copy of the GNU General Public License     **
**  along with this program.  If not, see http://www.gnu.org/licenses/.   **
**                                                                        **
****************************************************************************
**           Author: Andreas Beinstingel                                  **
**  Website/Contact: http://www.andreas-beinstingel.de                    **
**             Date: 07.12.15                                             **
**          Version: 1.0                                                  **
****************************************************************************/

#include "hmFem.h"
#include <malloc.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>
#include <iostream>
#include <math.h>
#include <stdio.h>      /* printf */
#include <vector>
using namespace std;
typedef vector<double> vec1d;
typedef vector<vec1d> vec2d;

    /*********************
     ****  M E M B E R ***
     *********************/
        //im header-file deklariert





    /************************************************
     ****  K O N -   U N D    D E S T R U K T O R ***
     ************************************************/
        hmFem::hmFem( linFunctions& linf ) : lins(linf)
        {



        }



        hmFem::~hmFem(){

        }


    /*********************
     ****  G E T T E R ***
     *********************/





    /*********************
     ****  S E T T E R ***
     *********************/
        void hmFem::setShowInfo( bool sInfo ) {
            this->showInfos = sInfo;
        }




    /***************************************
     ****  P U B L I C     M E T H O D S ***
     ***************************************/

        /* Loest das Sturm-Liouville-Problem      (adaptiv, nur linear)     *
         *                                                                  *
         *  -p(x)*u''(x)   +   q(x)*u'(x)   +   r(x)*u(x)   =   f(x)        *
         *                                                                  *
         * und gibt das Ergebnis in einem zweispaltigen Vektor zurueck      *
         * sol[:][0] = xWerte                                               *
         * sol[:][1] = yWerte = u Verschiebungsvektor                       *
         * sol[0][2] = erreichtes DurchSchnittsEta (=Fehler)                */
        vec2d hmFem::fem1d_sturmLiouvilleAdp(double a, double b, int bcType[2], double bcVal[2]) {
            vec1d eta, xVec;
            vec2d sl, rtrn;
            vector<vec2d> sly;
            double  hlp, ay, by, tol, etaAv, currEta, bcValY[2],
                    ur, ul, uyl, uyr, xl, xr, xyl, xyr, hy, up, uyp, integral, fa, fm, fb, t, ty,
                    intTMP;
            int N, counter, jl, jr, ny, bcTypeY[2];
            bool loop = true;
            bool showAdpInfo = false;
            etaAv = 0;

            //Eigene Info's, sonst zu viel
            if ( this->showInfos == true ) {
                showAdpInfo = true;
                this->showInfos = false;
            }

            tol = 0.0001;
            N = 4;

            if ( a > b) {
                hlp = a;
                a = b;
                b = hlp;
            }



            //Eine Loesung mit N Elementen generieren
            sl = this->fem1d_sturmLiouville( a, b, 1, N, bcType, bcVal);


            //Unbekannter Fehler, evtl. in der Definition des eigtl. Problems
            if (sl.size() < 1 ) {
                if (showAdpInfo) cerr << "Unbekannter Fehler im 1. Durchlauf..." << endl;
                return sl;
            }


            /* Ueber jeden Knoten eine neue Detail-Loesung          *
             * berechnen mit den Grenzen:                           *
             * ay = x(j-1)      und     by = x(j+1)                 *
             *                                                      *
             * Als Randwerte zaehlen hier die Ergebnisse der        *
             * vorherigen Loesung als Dirichlet-Bedingung:          *
             * uy(ay) = u(j-1)  und     uy(by) = u(j+1)             *
             *                                                      *
             * Anzahl der Unterlemente in diesem Intervall = 4      *
             * (ausser Erstes und Letztes Intervall, hier: 2)       *
             *                                                      *
             * Knoten 1 und Knoten N+1 sind Sonderfaelle!!          */

            // Randbedingungen immer als Dirichlet beachten,
            // da die Verschiebungen "gegeben" sind
            bcTypeY[0] = 0;     bcTypeY[1] = 0;

            counter = 1;
            while ( loop == true ) { //Schleife bis toleranz erreicht wurde



                /*  Schleife ueber alle "alten" Knoten und  *
                 *  lokale Sub-Probleme berechnen           */
                for ( unsigned int j=0; j<sl.size(); j++ ) {
                    sly.push_back( vec2d() );
                    if ( j == 0 ) {
                        ny = 2;
                        jl = 0;     jr = j+1;
                    }
                    else if (j == (sl.size()-1) ) {
                        ny = 2;
                        jl = j-1;   jr = j;
                    }
                    else {
                        ny = 4;
                        jl = j-1;   jr = j+1;
                    }


                    //Sub-intervall Grenzen setzen
                    ay = sl[jl][0];         by = sl[jr][0];

                    //RB setzen
                    bcValY[0] = sl[jl][1];  bcValY[1] = sl[jr][1];

                    //Sub-Problem berechnen
                    sly[j] = this->fem1d_sturmLiouville( ay, by, 1, ny, bcTypeY, bcValY);
                }




                /* Schleife ueber alle "alten" Elemente -> eta Fehler ueber H1-Norm             *
                 *  ->  doppelte Fehlerabschaetzung                                             *
                 *  ->  zB Element 3:    Fehler zw. rechten subIntervall von Knoten 2 und E 3   *
                 *                      +Fehler zw. linken  subIntervall von Knoten 3 und E 3   *
                 *      Erstes und Letztes Element = Sonderfall)                                *
                 *                                                                              *
                 *  pro Intervall:                                                              *
                 *  SQRT(       INT( (u - uy)^2 + (u' - uy')^2 )   )    in (ay, by)             *
                 *                                                                              *
                 *  u = vorherige Loesung (1E)  uy = neue Loesung (2 E)                         *
                 *  u'= Steigung des lin. E     uy'= jweilige Steigung des lin. E               *
                 *                                                                              *
                 *  Einfache Integration mit Simpson - Grobes Prinzip:                          *
                 *                                                                              *
                 *  3 Stuetzstellen ueber das Intervall 0 bis 1:                                *
                 *      x-Pos:          0   |   0.5 |   1                                       *
                 *      Gewichtung:     1   |   2   |   1                                       *
                 *      Q = h/6 * (fa * 4fm + fb)                                               *
                 *                                                                              */
                etaAv = 0;
                eta.clear();
                eta.push_back( 0.0 ); // 0-Element -> schoenes Schleifeninkrement
                for ( unsigned int e=1; e<sl.size(); e++  ) {

                    currEta = 0;

                    //aktuelle Elementdaten
                    xl = sl[e-1][0];
                    xr = sl[e][0];
                    ul = sl[e-1][1];
                    ur = sl[e][1];

                    //Numerisch zu kleiner Abstand -> bald teilen durch 0
                    if ( fabs(xr - xl) <= 0.00001 ) {
                        if (showAdpInfo) cerr << "Abbruch: Zu kleine Schritte werden inzwischen benoetigt!" << endl;
                        break;
                    }

                    up = (ur - ul) / (xr - xl);
                    t  = ur - up*xr;    //fuer Geradengleichung, quasi als Funktion

                    integral = 0.0;
                    //Rechte Seite von linken Knoten
                    if ( e < (sl.size()-1) ) {

                        //Schleife ueber die 2 Unterelemente
                        for ( int ey=1; ey<=2; ey++) {
                            //Daten des Vergleichsintervalls
                            if ( e == 1 ) {
                                xyl = sly[e-1][ey-1][0];
                                xyr = sly[e-1][ey][0];
                                uyl = sly[e-1][ey-1][1];
                                uyr = sly[e-1][ey][1];
                            } else {
                                xyl = sly[e-1][ey+1][0];
                                xyr = sly[e-1][ey+2][0];
                                uyl = sly[e-1][ey+1][1];
                                uyr = sly[e-1][ey+2][1];
                            }
                            uyp = ( uyr - uyl) / (xyr - xyl);
                            hy  = fabs(xyr - xyl);
                            ty  = uyr - uyp*xyr; //fuer Geradengleichung, quasi als Funktion

                            //Simpson Integration
                            // -> (Funktionen sind Geraden bei linearen Ansatzfunktionen -> y = mx + t)
                            fa = (up* xyl        +t) - uyl;
                            fm = (up*(xyl+(hy/2))+t) - (uyp*(xyl+(hy/2))+ty);
                            fb = (up*(xyl+hy)    +t) - uyr;

                            intTMP = 0;
                            intTMP = intTMP +  pow(fa,2) + pow((up - uyp),2);
                            intTMP = intTMP + (pow(fm,2) + pow((up - uyp),2))*4;
                            intTMP = intTMP +  pow(fb,2) + pow((up - uyp),2);
                            intTMP = intTMP * (hy/6);
                            integral = integral + intTMP;
                        }
                    }
                    currEta = currEta + 0.5*sqrt(integral);

                    integral = 0.0;
                    //Linke Seite von rechten Knoten
                    if ( e > 1 ) {
                        //Schleife ueber die 2 Unterelemente
                        for ( int ey=1; ey<=2; ey++) {
                            //Daten des Vergleichsintervalls
                            xyl = sly[e][ey-1][0];
                            xyr = sly[e][ey][0];
                            uyl = sly[e][ey-1][1];
                            uyr = sly[e][ey][1];

                            uyp = ( uyr - uyl) / (xyr - xyl);
                            hy  = fabs(xyr - xyl);
                            ty  = uyr - uyp*xyr; //fuer Geradengleichung, quasi als Funktion

                            //Simpson Integration
                            // -> (Funktionen sind Geraden bei linearen Ansatzfunktionen -> y = mx + t)
                            fa = (up* xyl        +t) - uyl;
                            fm = (up*(xyl+(hy/2))+t) - (uyp*(xyl+(hy/2))+ty);
                            fb = (up*(xyl+hy)    +t) - uyr;

                            intTMP = 0;
                            intTMP = intTMP +  pow(fa,2) + pow((up - uyp),2);
                            intTMP = intTMP + (pow(fm,2) + pow((up - uyp),2))*4;
                            intTMP = intTMP +  pow(fb,2) + pow((up - uyp),2);
                            intTMP = intTMP * (hy/6);
                            integral = integral + intTMP;
                        }

                    }
                    currEta = currEta + 0.5*sqrt(integral);
                    eta.push_back( currEta );
                    etaAv = etaAv + currEta;
                }
                etaAv = etaAv / (eta.size()-1);



                /* Setze einen Vektor der die x-Positionen fuer einen neuen Loesungsvorschlag
                 * beinhaltet und berechne die neue Loesung
                 * -> Es kommen ueberall neue Knoten hinzu, wo ETA zu gross ist       */
                loop = false;
                xVec.clear();
                for ( unsigned int e=1; e<sl.size(); e++) {
                    if ( eta[e] >  1.0  /* max( 1.0*etaAv , (  tol / (sl.size()-1) ) ) */ ) {
                        if ( e == 1 ) {
                            xVec.push_back( sl[e-1][0] );
                        }
                        xVec.push_back( sl[e-1][0] + ( fabs(sl[e][0] - sl[e-1][0]) * 1/ 4) );
                        xVec.push_back( sl[e-1][0] + ( fabs(sl[e][0] - sl[e-1][0]) * 2/ 4) );
                        xVec.push_back( sl[e-1][0] + ( fabs(sl[e][0] - sl[e-1][0]) * 3/ 4) );
                        xVec.push_back( sl[e][0] );
                        loop = true;
                        if (showAdpInfo) {
                            cout << counter <<". Iteration und verfeinere Element: " << e;
                            cout << " in 4 Unterelemente" << endl;
                        }
                    }
                    else if ( eta[e] <= 1.0 && eta[e] > 0.5 ) {
                        if ( e == 1 ) {
                            xVec.push_back( sl[e-1][0] );
                        }
                        xVec.push_back( sl[e-1][0] + ( fabs(sl[e][0] - sl[e-1][0]) * 1/ 3) );
                        xVec.push_back( sl[e-1][0] + ( fabs(sl[e][0] - sl[e-1][0]) * 2/ 3) );
                        xVec.push_back( sl[e][0] );
                        loop = true;
                        if (showAdpInfo) {
                            cout << counter <<". Iteration und verfeinere Element: " << e;
                            cout << " in 3 Unterelemente" << endl;
                        }
                    }
                    else if ( eta[e] <= 0.5 && eta[e] > 0.05  ) {
                        if ( e == 1 ) {
                            xVec.push_back( sl[e-1][0] );
                        }
                        xVec.push_back( sl[e-1][0] + ( fabs(sl[e][0] - sl[e-1][0]) * 1/ 2) );
                        xVec.push_back( sl[e][0] );
                        loop = true;
                        if (showAdpInfo) {
                            cout << counter <<". Iteration und verfeinere Element: " << e;
                            cout << " in 2 Unterelemente" << endl;
                        }
                    }
                    else {
                        if ( e == 1 ) {
                            xVec.push_back( sl[e-1][0] );
                        }
                        xVec.push_back( sl[e][0] );
                    }
                }

                //Falls noch ein Durchlauf benoetigt wird -> rechne die neue "alte" Loesung
                //ansonsten diesen Schritt sparen
                if ( loop == true ) {
                    if ( xVec.size() < 500 ) {
                        sl.clear();
                        sl = this->fem1d_sturmLiouville( a, b, 1, 1, bcType, bcVal, xVec);

                        //wieder auf Null fuer die naechste Runde
                        sly.clear();
                        eta.clear();

                        counter = counter + 1;
                    }
                    else {
                        loop = false;
                    }
                }


                //Vor Dauerschleife schuetzen
                if ( sl.size() > 500 ) {
                    loop = false;
                }
            }


            if ( showAdpInfo == true ) {
                this->showInfos = true;
            }
            if (showAdpInfo) cout << "Erreichtes DurchschnittsETA: " << etaAv << endl;
            if (showAdpInfo) cout << "Prozess fertig." << endl;

            for ( unsigned int i=0; i<sl.size(); i++ ) {
                rtrn.push_back(  vec1d() );
                rtrn[i].push_back(  sl[i][0] );
                rtrn[i].push_back(  sl[i][1] );
                rtrn[i].push_back(  etaAv );
            }
            return rtrn;


        }




        /* Loest das Sturm-Liouville-Problem      (NICHT adaptiv!!!)        *
         *                                                                  *
         *  -p(x)*u''(x)   +   q(x)*u'(x)   +   r(x)*u(x)   =   f(x)        *
         *                                                                  *
         * und gibt das Ergebnis in einem zweispaltigen Vektor zurueck      *
         * sol[:][0] = xWerte                                               *
         * sol[:][1] = yWerte = u Verschiebungsvektor                       */
        vec2d hmFem::fem1d_sturmLiouville(double a, double b, int ffeOrder, int N, int bcType[2], double bcVal[2], const vec1d &xVals ) {
            vec2d rtrn;     //Rueckgabematrix
            vec1d x;        //Positionsvektor x
            vec1d u;        //Verschiebungsvektor u (=gesuchte Loesung)
            vec2d K;        //Steifigkeitsmatrix (global)
            vec2d K_l;      //Steifigkeitsmatrix (lokal)
            vec2d M_l;      //Massenmatrix (lokal) (nur fuer die Eigenwertwertberechnung notwendig)
            vec2d M;        //Massenmatrix (global)(nur fuer die Eigenwertwertberechnung notwendig)
            vec1d f_l;      //Lastvektor (lokal)
            vec1d f;        //Lastvektor (global)
            int DOF_l;      //Freiheitsgrade pro Element = Knoten (lokal)
            int DOF_g;      //Anzahl der Knoten (Gesamt)
            double tmp;

            //xVals nur fuer lineare Ansatzfunktionen!!!
            if ( xVals.size() > 0 && ffeOrder == 1 ) {
                N = xVals.size() - 1;
            }

            if (this->showInfos) cout << "Eingaben verarbeiten.. " << endl;
            this->feOrder   = ffeOrder;
            DOF_l           =  this->feOrder + 1;
            DOF_g           = (this->feOrder * N) + 1;



            if ( a == b) {
                return rtrn; //kein Intervall gegeben...
            }

            if ( a > b) {                
                tmp = a;
                a = b;
                b = tmp;
            }
            this->h         = (b - a) / (N*this->feOrder);

            //interne Knotengrenze!
            if ( DOF_g < 2 || DOF_g > 500) {
                return rtrn;
            }

            //Matrizen und Vektoren auf die richtige Groesse bringen und mit "0" initialisieren
                //Globale Matrizen | Vektoren
                for (  int i=0; i< DOF_g ; i++) {
                    rtrn.push_back(  vec1d() );
                    K.push_back(  vec1d() );
                    M.push_back(  vec1d() );
                    if ( xVals.size() > 0 && ffeOrder == 1 ) {
                        x.push_back(  xVals[i] );
                    }
                    else {
                        this->h         = (b - a) / (N*this->feOrder);
                        x.push_back(  a+(i*h) );    //x-Positionen abspeichern (aequidistant)
                    }
                    u.push_back(  0.0 );
                    f.push_back(  0.0 );
                    for (  int j=0; j< DOF_g ; j++) {
                         K[i].push_back(  0.0 );
                         M[i].push_back(  0.0 );
                         if ( j < 2) {
                            rtrn[i].push_back(  0.0 );
                         }
                    }
                }
                //lokale Matrizen | Vektoren
                for (int i=0; i<DOF_l; i++) {
                    K_l.push_back(  vec1d() );
                    M_l.push_back(  vec1d() );
                    f_l.push_back(  0.0 );
                    for (int j=0; j<DOF_l; j++) {
                        K_l[i].push_back(  0.0 );
                        M_l[i].push_back(  0.0 );
                    }
                }


            //Matrizen und Vektoren bzgl. des Problems berechnen
                /* Schleife ueber alle Elemente;
                 * pro Element lokales Feld berechnen und in das globale einordnen
                 * lokale Felder immer neu integrieren, da hier lineare Koeffizienten definiert werden koennen  */
                if (this->showInfos) cout << "Steifigkeitsmatrix K und Lastvektor f aufbauen " << endl;
                for ( int e = 1; e <= N; e++ ) {
                    if ( xVals.size() > 0  && ffeOrder == 1 ) {
                        this->xim1  = xVals[e-1];
                        this->xi    = xVals[e];
                    }
                    else {
                        this->xim1  = a+((e-1)  *  (b-a)/N  );
                        this->xi    = a+( e     *  (b-a)/N  );
                    }

                    this->h     = fabs( this->xi - this->xim1 );

                    K_l = this->getK_loc( &hmFem::LHS_sturmLiouville );
                    K   = this->addToGSM( K_l, e, K  );

                    M_l = this->getM_loc(  );
                    M   = this->addToGSM( M_l, e, M  );

                    f_l = this->getF_loc(  );
                    f   = this->addToGSV( f_l, e, f  );
                }

                if (this->showInfos) cout << "Randbedingungen setzen " << endl;
                //Randbedingungen einbauen
                    // Dirichlet
                    if (bcType[0] == 0) {
                        f[0] = bcVal[0];
                        f[1] = f[1] - bcVal[0] * K[1][0];
                        if (DOF_g > 2) {
                        f[2] = f[2] - bcVal[0] * K[2][0];
                        }
                        if (DOF_g > 3) {
                        f[3] = f[3] - bcVal[0] * K[3][0];
                        }
                        for ( int i=0; i<DOF_g; i++) {
                            K[0][i] = 0.0; K[i][0] = 0.0;
                        }
                        K[0][0] = 1;
                    }
                    if (bcType[1] == 0) {
                        f[DOF_g-1] = bcVal[1];
                        f[DOF_g-2] = f[DOF_g-2] - bcVal[1] * K[DOF_g-2][DOF_g-1];
                        if (DOF_g > 2) {
                        f[DOF_g-3] = f[DOF_g-3] - bcVal[1] * K[DOF_g-3][DOF_g-1];
                        }
                        if (DOF_g > 3) {
                        f[DOF_g-4] = f[DOF_g-4] - bcVal[1] * K[DOF_g-4][DOF_g-1];
                        }
                        for ( int i=0; i<DOF_g; i++) {
                            K[DOF_g-1][i] = 0.0; K[i][DOF_g-1] = 0.0;
                        }
                        K[DOF_g-1][DOF_g-1] = 1;

                    }
                    // Neumann
                    if (bcType[0] == 1) {
                        f[0] = f[0] - (this->lins.funPx(  (a*this->h)+this->xim1  )*bcVal[0]);
                    }
                    if (bcType[1] == 1) {
                        f[DOF_g-1] = f[DOF_g-1] + (this->lins.funPx(  (b*this->h)+this->xim1  )*bcVal[1]);
                    }

                //Verschiebungsvektor berechnen
                    u = this->GLS( K, f);

                //Eigenwerte berechnen

                    //TODO: this->eig(  this->inv(M)*K  );
                    /*
                     * Achtung: hierfuer muss die linke Seite in Normalform vorliegen
                     *          -> p(x) = 1, also kein Vorfaktor bei u''
                     *          -> man koennte auch
                     *          -> q(x) / p(x) und r(x)/p(x) und f(x)/p(x) setzen,
                     *                  aber Vorsicht: durch NULL teilen tut man nicht oO
                     */

                if (this->showInfos) cout << "Loesungen zurueckgeben " << endl;
                //Ausgabematrix belegen
                    for ( int i=0; i<DOF_g; i++ ) {
                        rtrn[i][0] = x[i];
                        rtrn[i][1] = u[i];
                    }
                    return rtrn;
        }





        /* Loest die Bernoulli Biege-DGL                                    *
         *                                                                  *
         *  -[E*I*u''(x)]''   =   q(x)    in (0,L)                          *
         *                                                                  *
         * und gibt das Ergebnis in einem zweispaltigen Vektor zurueck      *
         * sol[:][0] = xWerte                                               *
         * sol[:][1] = yWerte = u Verschiebungsvektor                       */
        vec2d hmFem::fem1d_bernoulliBeam(double qVal, double qPos, double E, double I, double L, int N, int bc[2][3]) {
            vec2d rtrn;     //Rueckgabematrix
            vec1d x;        //Positionsvektor x
            vec1d u;        //Verschiebungsvektor u (=gesuchte Loesung)
            vec2d K;        //Steifigkeitsmatrix (global)
            vec2d K_l;      //Steifigkeitsmatrix (lokal)
            vec1d f_l;      //Lastvektor (lokal)
            vec1d f;        //Lastvektor (global)
            int DOF_l;      //Freiheitsgrade pro Element = Knoten (lokal)
            int DOF_g;      //Anzahl der Knoten (Gesamt)
            double a;
            double b;
            double tmp;

            if (this->showInfos) cout << "Eingabe verarbeiten " << endl;
            this->feOrder   =   3;                          //min. 3 wegen 2. Ableitung in Variationsformulierung
            DOF_l           =   this->feOrder + 1;
            DOF_g           =   (N+1) * (this->feOrder-1);  //Doppelte Ueberlappung der K_l in K


            if ( L <= 0) {
                if (this->showInfos) cerr << "Negative Balkenlaenge " << endl;
                return rtrn;    //kein Intervall gegeben...
            }

            if ( qPos > L) {
                if (this->showInfos) cerr << "Kraft greift ausserhalb des Balkens an " << endl;
                return rtrn;    //Kraft ausserhalb des Balkens..
            }

            a = 0;
            b = L;
            this->h     = (b - a) / N;

            //interne Knotengrenze! (siehe Randbedingungen!)
            if ( DOF_g < 4 || DOF_g > 500) {
                if (this->showInfos) cerr << "Anzahl der Knoten sollte zwischen 4 und 500 liegen. " << endl;
                return rtrn;
            }

            //Matrizen und Vektoren auf die richtige Groesse bringen und mit "0" initialisieren
                if (this->showInfos) cout << "Steifigkeitsmtarix K und Lastvektor f aufbauen " << endl;
                //Globale Matrizen | Vektoren
                int ci = 0;
                for (  int i=0; i< DOF_g ; i++) {
                    rtrn.push_back(  vec1d() );
                    K.push_back(  vec1d() );
                    if ( i % 2 == 0 && i != 0  ) {
                        ci = ci + 1;
                    }
                    x.push_back(  a+(ci*this->h) );    //x-Positionen abspeichern (aequidistant)
                    u.push_back(  0.0 );
                    f.push_back(  0.0 );
                    for (  int j=0; j< DOF_g ; j++) {
                         K[i].push_back(  0.0 );
                         if ( j < 2) {
                            rtrn[i].push_back(  0.0 );
                         }
                    }
                }
                //lokale Matrizen | Vektoren
                for (int i=0; i<DOF_l; i++) {
                    K_l.push_back(  vec1d() );
                    f_l.push_back(  0.0 );
                    for (int j=0; j<DOF_l; j++) {
                        K_l[i].push_back(  0.0 );
                    }
                }


            //Steifigkeitsmatrix K_l berechnen und in K einsetzen (1 Knoten hat u und u' !!!)
                for ( int e = 1; e <= N; e++ ) {
                    this->xim1  = a+((e-1)  *  (b-a)/N  );
                    this->xi    = a+( e     *  (b-a)/N  );
                    this->h     = fabs( this->xi - this->xim1 );

                    /*  Statisch die ESM fuer den Bernoulli-Balken belegen
                     *  Hier werden Hermitische Polynome als Ansatzfunktionen genutzt,
                     *  so koenne auch die Winkel/Steigungen an den Knotenpunkten beruecksichtigt
                     *  werden.
                     * */
                    tmp = ((E*I)/(this->h*this->h*this->h));
                    K_l[0][0] = tmp * (+12);
                    K_l[0][1] = tmp * (-6 * this->h);
                    K_l[0][2] = tmp * (-12);
                    K_l[0][3] = tmp * (-6 * this->h);

                    K_l[1][0] = tmp * (-6 * this->h);
                    K_l[1][1] = tmp * (+4 * this->h*this->h);
                    K_l[1][2] = tmp * (+6 * this->h);
                    K_l[1][3] = tmp * (+2 * this->h*this->h);

                    K_l[2][0] = tmp * (-12);
                    K_l[2][1] = tmp * (+6 * this->h);
                    K_l[2][2] = tmp * (+12);
                    K_l[2][3] = tmp * (+6 * this->h);

                    K_l[3][0] = tmp * (-6 * this->h);
                    K_l[3][1] = tmp * (+2 * this->h*this->h);
                    K_l[3][2] = tmp * (+6 * this->h);
                    K_l[3][3] = tmp * (+4 * this->h*this->h);

                    this->feOrder = 2; /* doppelte Uerblappung wegen Winkel  */
                    K   = this->addToGSM( K_l, e, K  );

                }

               //Punktlast aufbringen -> Last auf den nächstgelegenen Knoten setzen
                   for (int i=0; i<(DOF_g-2); i = i + 2) {
                        if ( qPos > x[i] && qPos <= x[i+2]   ) {
                            if (  fabs(qPos - x[i]) <=  fabs(qPos - x[i+2])     ) {
                                f[i] = qVal;
                            }
                            else {
                                f[i+2] = qVal;
                            }
                        }
                   }

                if (this->showInfos) cout << "Randbedingungen setzen " << endl;
                //Randbedingungen
                    // links
                    for ( int i=0; i<DOF_g; i++) {
                        if ( bc[0][1] == 1) {
                            K[0][i] = 0.0; K[i][0] = 0.0;
                            f[0] = 0.0;
                            K[0][0] = 1;
                        }
                        if ( bc[0][2] == 1) {
                            K[1][i] = 0.0; K[i][1] = 0.0;
                            f[1] = 0.0;
                            K[1][1] = 1;
                        }
                    }
                    //rechts
                        for ( int i=0; i<DOF_g; i++) {
                            if ( bc[1][1] == 1 ) {
                                K[DOF_g-2][i] = 0.0; K[i][DOF_g-2] = 0.0;
                                f[DOF_g-2] = 0.0;
                                K[DOF_g-2][DOF_g-2] = 1;
                            }
                            if ( bc[1][2] == 1) {
                                K[DOF_g-1][i] = 0.0; K[i][DOF_g-1] = 0.0;
                                f[DOF_g-1] = 0.0;
                                K[DOF_g-1][DOF_g-1] = 1;
                            }

                        }

                //Verschiebungsvektor berechnen
                    u = this->GLS( K, f);

                //Ausgabematrix belegen
                    if (u.size() > 0) {
                        for ( int i=0; i<DOF_g; i++ ) {
                            rtrn[i][0] = x[i];
                            rtrn[i][1] = u[i];
                        }
                    }
                    return rtrn;
        }




    /*****************************************
     ****  P R I V A T E     M E T H O D S ***
     *****************************************/


        /* Ansatzfunktionen abgeleitet aus den      *
         * Lagrange'schen Basispolynomen fuer 1d    */
        double hmFem::phi( double x, int node, int der, int order ) {

            //linear
            if ( order == 1 ) {
                if ( der == 0 ) {
                    if ( node == 1 ) {
                        return (1 - x);
                    }
                    else if ( node == 2 ) {
                        return x;
                    }
                    else {
                        return 0; //sollte nicht vorkommen
                    }
                }
                else if ( der == 1) {
                    if ( node == 1 ) {
                        return -1;
                    }
                    else if ( node == 2 ) {
                        return 1;
                    }
                    else {
                        return 0; //sollte nicht vorkommen
                    }
                }
                else {
                    return 0;
                }
            }
            //quadratisch
            else if ( order == 2 ) {
                if ( der == 0 ) {
                    if ( node == 1 ) {
                        return (+2*x*x - 3*x + 1);
                    }
                    else if ( node == 2 ) {
                        return (-4*x*x + 4*x);
                    }
                    else if ( node == 3 ) {
                        return (+2*x*x - 1*x);
                    }
                    else {
                        return 0; //sollte nicht vorkommen
                    }
                }
                else if ( der == 1)  {
                    if ( node == 1 ) {
                        return (+4*x - 3);
                    }
                    else if ( node == 2 ) {
                        return (-8*x + 4);
                    }
                    else if ( node == 3 ) {
                        return (+4*x - 1);
                    }
                    else {
                        return 0; //sollte nicht vorkommen
                    }
                }
                else if ( der == 2) {
                    if ( node == 1 ) {
                        return (+4);
                    }
                    else if ( node == 2 ) {
                        return (-8);
                    }
                    else if ( node == 3 ) {
                        return (+4);
                    }
                    else {
                        return 0; //sollte nicht vorkommen
                    }
                }
                else {
                    return 0;
                }
            }
            //kubisch
            else if ( order == 3 ) {
                if ( der == 0 ) {
                    if ( node == 1 ) {
                        return (-4.5*x*x*x + 9.0*x*x - 5.5*x + 1.0);
                    }
                    else if ( node == 2 ) {
                        return (+13.5*x*x*x - 22.5*x*x + 9.0*x);
                    }
                    else if ( node == 3 ) {
                        return (-13.5*x*x*x + 18.0*x*x - 4.5*x);
                    }
                    else if ( node == 4 ) {
                        return (+4.5*x*x*x - 4.5*x*x + 1.0*x);
                    }
                    else {
                        return 0; //sollte nicht vorkommen
                    }
                }
                else if ( der == 1)  {
                    if ( node == 1 ) {
                        return (-13.5*x*x + 18.0*x - 5.5);
                    }
                    else if ( node == 2 ) {
                        return (+40.5*x*x - 45.0*x + 9);
                    }
                    else if ( node == 3 ) {
                        return (-40.5*x*x + 36.0*x - 4.5);
                    }
                    else if ( node == 4 ) {
                        return (+13.5*x*x - 9.0*x + 1);
                    }
                    else {
                        return 0; //sollte nicht vorkommen
                    }
                }
                else if ( der == 2) {
                    if ( node == 1 ) {
                        return (-27.0*x + 18.0);
                    }
                    else if ( node == 2 ) {
                        return (+81.0*x - 45.0);
                    }
                    else if ( node == 3 ) {
                        return (-81.0*x + 36.0);
                    }
                    else if ( node == 4 ) {
                        return (+27.0*x - 9.0);
                    }
                    else {
                        return 0; //sollte nicht vorkommen
                    }
                }
                else if ( der == 3) {
                    if ( node == 1 ) {
                        return (-27.0);
                    }
                    else if ( node == 2 ) {
                        return (+81.0);
                    }
                    else if ( node == 3 ) {
                        return (-81.0);
                    }
                    else if ( node == 4 ) {
                        return (+27.0);
                    }
                    else {
                        return 0; //sollte nicht vorkommen
                    }
                }
                else {
                    return 0;
                }
            }
            else {
                return 0; //sollte nicht vorkommen
            }

        }



        //numerische Integration (Romberg-Verfahren)
        double hmFem::quadR( double (hmFem::*f)(double x), double a, double b, double eps = 0.0001, int maxIt = 30) {
            double T[100][100], h1, q;
            double rtrn;
            int  i, j ;
            bool loop;
            rtrn    = -999;

            i = 2;
            j = 2;
            h1 = (b-a);
            T[1][1] = ( h1 ) *  ( ( (this->*f)(a)+(this->*f)(b) )/2 );
            loop = true;
            while ( loop == true) {

                if ( i <= maxIt ) {
                    //Neue Schrittbreite
                    h1 = (b-a) / ( pow(2, (i-1)) );

                    //Naechsten Wert in der 1. Spalte berechnen
                    //k+2 -> Alte Werte gemerkt!!!
                    q = 0;
                    for ( int k=1; k <= ( pow(2, (i-1)) - 1 )  ; ) {
                        q = q + (this->*f)(a + k*h1);
                        k = k + 2;
                    }
                    T[i][1] = (T[i-1][1] / 2) + h1*q;


                    //Neue Zeile komplett "durch-extrapolieren"
                    for (int s=2; s <= j ; s++) {
                         T[i][s] = T[i][s-1] + (  ( T[i][s-1] - T[i-1][s-1] ) /  ( pow(2, (2*(s-1)))  - 1)  );
                    }
                    rtrn = T[i][j];


                    //Wurde die Genauigkeit jetzt erfuellt ??
                    if ( fabs(T[i-1][j-1] - T[i][j-1])  <= eps ) {
                        rtrn = T[i][j];
                        loop = false;
                    }
                    else {
                        i = i + 1;
                        j = j + 1;
                    }
                }
                else {
                    loop = false;
                }
            }

            return rtrn;
        }



        /* Loest ein lineares GLS: A * u  =  b
         * wenn A symm.     -> cholesky
         * wenn A unsymm.   -> gauss-elimination
         */
        vec1d hmFem::GLS( vec2d A, vec1d b) {
            vec1d u;

            //Matrix singular -> GLS nicht loesbar
            if (this->det(A) == 0 ) {
                if (this->showInfos) cerr << "Matrix singulaer! -> GLS nicht loesbar" << endl;
                return u;
            }


            if (  this->isSymm(A)  ) {
                if (this->showInfos) cout << "GLS aufloesen (Cholesky) " << endl;
                u = this->gaussE(A, b); //positiv definit wird nicht geprueft TODO
            }
            else if ( this->isQuadr(A)  )  {
                if (this->showInfos) cout << "GLS aufloesen (Gauss) " << endl;
                u = this->gaussE(A, b);
            }
            else {
                return u;
            }
            return u;
        }





        /* Loesen eines linearen GLS                    *
         * der Form: A * u = b (Cholesky - Zerlegung)   *
         *                                              *
         * A muss symmetrisch und positiv definit sein! */
        vec1d hmFem::chol( vec2d A, vec1d b) {
            vec1d u;
            vec1d y; //Hilfvektor
            vec2d R; //Hilfsmatrix
            int N;
            double summe;
            N = A.size();

            //Matrix A muss quadratisch und symmetrisch sein
                //-> wird schon in GLS(A, y) geprueft

            // Laenge des Vektors b muss gleich Matrix A Dimension sein
            if ( A.size() != b.size() ) {
                return u;
            }

            //Vektor u auf richtige Laenge setzen und default 0.0
            for (  int i=0; i < N ; i++     ) {
                u.push_back(0.0);
                y.push_back(0.0);       //Hilfsvektor gleich mit auf Null setzen
            }

            //Hilfsmatrix R auf Null setzen
            for (  int i=0; i<N ; i++) {
                R.push_back(  vec1d() );
                for (  int j=0; j<N ; j++) {
                     R[i].push_back(  0.0 );
                }
            }

            //Start der Cholesky Zerlegung
            for (  int i=0 ; i < N ; i++) {
                for (  int j=0 ; j < N ; j++ ) {
                    summe = 0;
                    for (  int k=0 ; k <= (i-1) ; k++ ) {
                        summe = summe +  ( R[k][i] * R[k][j] );
                    }
                    if ( i == j) {
                        R[i][i] = sqrt( A[i][i] - summe );
                    }
                    else {
                        if ( R[i][i] == 0) {
                                R[i][j] = 0;
                        }
                        else {
                                R[i][j] = ( A[i][j] - summe ) / R[i][i];
                        }
                    }
                }
            }

            //Vorwaerts einsetzen
            for ( int i=0 ; i < N ; i++) {
                summe = 0;
                for (  int k=0 ; k <= (i-1) ; k++ ) {
                    summe = summe + (  ( R[k][i]  * y[k] ) );
                }
                if ( R[i][i]  == 0) {
                    y[i] = 0;
                }
                else {
                     y[i] = (b[i] - summe) / R[i][i]  ;
                }

            }
            //Rueckwaerts einsetzen
            for ( int i=(N-1) ; i >= 0 ; i--) {
                summe = 0;
                for ( int k=(N-1) ; k >= (i+1) ; k-- ) {
                    summe = summe + (  ( R[i][k]  * u[k] ) );
                }
                if ( R[i][i] == 0) {
                    u[i] = 0;
                }
                else {
                    u[i] = (y[i] - summe)  / R[i][i];
                }
            }
            return u;
        }




        /* Loesen eines linearen GLS                    *
         * der Form: A * u = b (LU - Zerlegung)   *
         *                                              */
        vec1d hmFem::gaussE( vec2d A, vec1d b) {
            vec1d u, y;
            vec2d L, U;
            double summe;
            int N;
            N = A.size();


            //Matrix muss quadratisch sein
                //-> wird schon in GLS(A) geprueft

            for (  int i=0; i<N ; i++) {
                L.push_back(  vec1d() );
                U.push_back(  vec1d() );
                y.push_back( 0.0 );
                u.push_back( 0.0 );
                for (  int j=0; j<N ; j++) {
                     L[i].push_back(  0.0 );
                     U[i].push_back(  0.0 );
                }
            }

            //LU-Zerlegung
            this->lu_decomp(A, L, U);


            //Vorwaerts einsetzen
            for ( int i=0 ; i < N ; i++) {
                summe = 0;
                for (  int k=0 ; k <= (i-1) ; k++ ) {
                    summe = summe + (  ( L[i][k]  * y[k] ) );
                }
                if ( L[i][i]  == 0) {
                    y[i] = 0;
                }
                else {
                     y[i] = (b[i] - summe) / L[i][i]  ;
                }

            }
            //Rueckwaerts einsetzen
            for ( int i=(N-1) ; i >= 0 ; i--) {
                summe = 0;
                for ( int k=(N-1) ; k >= (i+1) ; k-- ) {
                    summe = summe + (  ( U[i][k]  * u[k] ) );
                }
                if ( U[i][i] == 0) {
                    u[i] = 0;
                }
                else {
                    u[i] = (y[i] - summe)  / U[i][i];
                }
            }


            return u;
        }



        /* Berechnet die Determinante von A     *
         *                                      *
         * A muss quadratisch sein!             */
        double hmFem::det( vec2d A) {
            double d = 0;
            double ll, uu;
            vec2d L, P, U;

            if ( !this->isQuadr(A) ) {
                return d;
            }

            for ( unsigned  int i=0; i<A.size() ; i++) {
                L.push_back(  vec1d() );
                U.push_back(  vec1d() );
                for ( unsigned int j=0; j<A.size() ; j++) {
                     L[i].push_back(  0.0 );
                     U[i].push_back(  0.0 );
                }
            }

            //LU-Zerlegung
            this->lu_decomp(A, L, U);

            ll = 1; uu = 1;
            for ( unsigned int i=0; i<A.size(); i++) {
                ll = ll * L[i][i]; //ll sollte immer 1 sein
                uu = uu * U[i][i];
            }
            d = ll * uu;


            return d;

        }



        /* Berechnet die Inverse von A      *
         *                                  *
         * A muss quadratisch sein!         */
        vec2d hmFem::inv(vec2d A) {
            vec2d invA;

            if ( !this->isQuadr(A) ) {
                return invA;
            }

            //TODO
            return invA;

        }



        /* Berechnet die Eigenwerte von A   *
         *                                  *
         * A muss quadratisch sein!         */
        vec1d hmFem::eig(vec2d A) {
            vec1d ew;

            if ( !this->isQuadr(A) ) {
                return ew;
            }

            //TODO
            return ew;
        }



        /* Prueft ob die Matrix A       *
         * quadratisch ist              */
        bool hmFem::isQuadr(vec2d A) {
            int N  = A.size();
            bool rtrn = true;
            for (  int i=0; i < N ; i++     ) {
                if ( A.size() != A[i].size() ) {
                    rtrn =  false;
                }
            }
            return rtrn;
        }


        /* Prueft ob die Matrix A       *
         * symmetrisch ist              */
        bool hmFem::isSymm(vec2d A) {
            int N  = A.size();
            bool rtrn = true;
            if ( isQuadr(A)  ) {
                for (  int i=0; i < N ; i++     ) {
                    for (  int j=0; j < N ; j++     ) {
                        if ( fabs(A[i][j] - A[j][i]) >= 0.0001 ) {
                            rtrn = false;
                        }
                    }
                }
            }
            else {
                rtrn =  false;
            }
            return rtrn;

        }


        /* Zerlegt die Matrix A in eine                 *
         * untere Dreiecksmatrix L und in               *
         * eine obere Dreiecksmatrix U ohne             *
         * Pivotisierung P                              *
         * Uebergaben in dem Fall per Reference         */
        void hmFem::lu_decomp(const vec2d &A, vec2d &L, vec2d &U) {
            double sum = 0;
            int N;
            N = A.size();

            for (int i = 0; i < N; i++) {
                U[i][i] = 1;
            }

            for (int j = 0; j < N; j++) {
                for (int i = j; i < N; i++) {
                    sum = 0;
                    for (int k = 0; k < j; k++) {
                        sum = sum + L[i][k] * U[k][j];
                    }
                    L[i][j] = A[i][j] - sum;
                }

                for (int i = j; i < N; i++) {
                    sum = 0;
                    for(int k = 0; k < j; k++) {
                        sum = sum + L[j][k] * U[k][i];
                    }
                    if (L[j][j] == 0) {
                        //teilen durch Null geht nicht oO
                        return;
                    }
                    U[j][i] = (A[j][i] - sum) / L[j][j];
                }
            }
        }




        /* Eine lokale SteifigkeitsMatrix wird in       *
         * eine globale SteifigkeitsMatrix eingebaut    */
        vec2d hmFem::addToGSM( vec2d K_l, int e, vec2d K ) {
            vec2d K_result;
            int DOF_l = K_l.size();
            for (int i=0; i<DOF_l; i++) {
                for (int j=0; j<DOF_l; j++) {
                    K[i+((e-1)*feOrder)][j+((e-1)*feOrder)] = K[i+((e-1)*feOrder)][j+((e-1)*feOrder)] + K_l[i][j];
                }
            }
            K_result = K;
            return K_result;
        }


        /* Ein lokaler SteifigkeitsVektor (=rechte Seite) wird in       *
         * einen globalen SteifigkeitsVektor eingebaut                  */
        vec1d hmFem::addToGSV( vec1d f_l, int e, vec1d f ) {
            vec1d f_result;
            for (int i=0; i<(feOrder+1); i++) {
                f[i+((e-1)*feOrder)] = f[i+((e-1)*feOrder)] + f_l[i];
            }
            f_result = f;
            return f_result;
        }




        /*  Aufbau einer lokalen Steifigkeitsmatrix         *
         *                                                  *
         *  linke Seite der Variationsformulierung          *
         *  kann individuell uebergeben werden              */
        vec2d hmFem::getK_loc( double (hmFem::*LHSvariationsForm)(double x) ) {
            vec2d K_l;

            //Erstmal auf Groesse setzen und mit "0" initialisieren
            for (int i=0; i<(feOrder+1); i++) {
                K_l.push_back(  vec1d() );
                for (int j=0; j<(feOrder+1); j++) {
                    K_l[i].push_back(  0.0 );
                }
            }

            //Ueber Variationsformulierung die einzelnen Elemente aufintegrieren
            for (int i=0; i<(feOrder+1); i++) {
                for (int j=0; j<(feOrder+1); j++) {
                    this->m = i+1;  //Array geht bei Null los, aber Ansatzfunktionen bei 1..
                    this->n = j+1;
                    K_l[i][j] = K_l[i][j] + this->quadR( LHSvariationsForm, 0, 1, 0.0001, 5);
                }
            }

            return K_l;
        }






        /* Integriert einen lokalen Lastvektor          *
         * (=rechte Seite d. Variationsformulierung)    *
         *                                              *
         *                                              */
        vec1d hmFem::getF_loc( ) {
            vec1d f_l;

            //Erstmal auf Groesse setzen und mit "0" initialisieren
            for (int i=0; i<(feOrder+1); i++) {
                f_l.push_back(  0.0 );
            }

            //Ueber RHS_f die einzelnen Elemente aufintegrieren
            for (int i=0; i<(feOrder+1); i++) {
                    this->m = i+1;  //Array geht bei Null los, aber Ansatzfunktion bei 1!
                    f_l[i] = f_l[i] + this->quadR( &hmFem::RHS_f, 0, 1, 0.0001, 5);
            }
            return f_l;
        }
        //Die zu integrierende 'Variationsformulierung' fuer den Lastvektor auf der rechten Seite
        double hmFem::RHS_f(double x) {
            double y;
            double f;
            f =  this->lins.funFx(  (x*this->h)+this->xim1   );
            y =  this->h * f * this->phi(x, this->m, 0, this->feOrder);
            return y;
        }








        /* Integriert die lokale Massenmatrix   *
         * (zB fuer Eigenwertproblem!)          *
         *                                      *
         * zwischen xim1 und xi                 */
        vec2d hmFem::getM_loc( ) {
            vec2d M_l;

            //Erstmal auf Groesse setzen und mit "0" initialisieren
            for (int i=0; i<(feOrder+1); i++) {
                M_l.push_back(  vec1d() );
                for (int j=0; j<(feOrder+1); j++) {
                    M_l[i].push_back(  0.0 );
                }
            }

            //Ueber Variationsformulierung die einzelnen Elemente aufintegrieren
            for (int i=0; i<(feOrder+1); i++) {
                for (int j=0; j<(feOrder+1); j++) {
                    this->m = i+1;  //Array geht bei Null los, aber Ansatzfunktionen bei 1...
                    this->n = j+1;
                    M_l[i][j] = M_l[i][j] + this->quadR( &hmFem::RHS_eig, 0, 1, 0.0001, 5);
                }
            }
            return M_l;
        }
        //Die zu integrierende 'Variationsformulierung' fuer die MassenMatrix (=rechte Seite: lambda*u(x))
        double hmFem::RHS_eig(double x) {
            double y;
            y =    this->h * this->phi(x, this->n, 0, this->feOrder) * this->phi(x, this->m, 0, this->feOrder);
            return y;
        }





        /*  Variationsformulierungen der linken Seite hier definieren       *
         *  ->  so entstehen verschiedene Steifigkeitsmatrizen              *
         *      "Elementbibliothek"                                         *
         *      fuer verschiedenen DGL Probleme                             */




        // Sturm-Liouville: -p(x)*u''(x) + q(x)*u'(x) + r(x)*u(x)
        double hmFem::LHS_sturmLiouville(double x) {
            double y;
            double q, r, p;

            p   =  this->lins.funPx(  (x*this->h)+this->xim1  );
            q   =  this->lins.funQx(  (x*this->h)+this->xim1  );
            r   =  this->lins.funRx(  (x*this->h)+this->xim1  );

            y =    (1/this->h) *  p * this->phi(x, this->n, 1, this->feOrder)    * this->phi(x, this->m, 1, this->feOrder) +
                                  q * this->phi(x, this->n, 1, this->feOrder)    * this->phi(x, this->m, 0, this->feOrder) +
                   this->h     *  r * this->phi(x, this->n, 0, this->feOrder)    * this->phi(x, this->m, 0, this->feOrder);

            return y;
        }














































