/***************************************************************************
**                                                                        **
**  hmFem - Klasse														  **
**	Numerische Loesung von 1D Randwertproblemen (RWP)					  **
**	mit Hilfe der Methode der Finiten Elemente  (FEM)					  **
**  Copyright (C) 2015 Andreas Beinstingel                                **
**                                                                        **
**  This program is free software: you can redistribute it and/or modify  **
**  it under the terms of the GNU General Public License as published by  **
**  the Free Software Foundation, either version 3 of the License, or     **
**  (at your option) any later version.                                   **
**                                                                        **
**  This program is distributed in the hope that it will be useful,       **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of        **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         **
**  GNU General Public License for more details.                          **
**                                                                        **
**  You should have received a copy of the GNU General Public License     **
**  along with this program.  If not, see http://www.gnu.org/licenses/.   **
**                                                                        **
****************************************************************************
**           Author: Andreas Beinstingel                                  **
**  Website/Contact: http://www.andreas-beinstingel.de                    **
**             Date: 07.12.15                                             **
**          Version: 1.0                                                  **
****************************************************************************/


#ifndef HMFEM_Included
#define HMFEM_Included

#include <iostream>
#include <math.h>
#include <stdio.h>      /* printf */
#include <stdlib.h>     /* abs */
#include <vector>

using namespace std;
typedef vector<double> vec1d;
typedef vector<vec1d> vec2d;


    /*  Mit Hilfe dieser Vaterklasse koennen
     *  von ausserhalb Funktionen definiert werden (=ueberladen!)
     *  in C++ muss das vorher schon ueber "virtual" gesetzt werden!
     *  -> MainWindow : public linFunctions
     */
    class linFunctions {
        public:
            virtual double funPx(double x) = 0;
            virtual double funQx(double x) = 0;
            virtual double funRx(double x) = 0;
            virtual double funFx(double x) = 0;
    };


    class hmFem {
		
        /*********************
         ****  M E M B E R ***
         *********************/
        private:
            linFunctions& lins;
            double xi, xim1, h;
            int feOrder, m, n;
            bool showInfos = true;




        /************************************************
         ****  K O N -   U N D    D E S T R U K T O R ***
         ************************************************/
        public:
            hmFem( linFunctions& linf );
            ~hmFem();


        /*********************
         ****  G E T T E R ***
         *********************/
        public:



        /*********************
         ****  S E T T E R ***
         *********************/
        public:
            void setShowInfo( bool sInfo );



        /***************************************
         ****  P U B L I C     M E T H O D S ***
         ***************************************/
        public:
            vec2d fem1d_sturmLiouvilleAdp(double a, double b, int bcType[2], double bcVal[2]);
            vec2d fem1d_sturmLiouville(double a, double b, int ffeOrder, int N, int bcType[2], double bcVal[2], const vec1d &xVals = vec1d() );
            vec2d fem1d_bernoulliBeam(double qVal, double qPos, double E, double I, double L, int numElements, int bc[2][3]);





        /*****************************************
         ****  P R I V A T E     M E T H O D S ***
         *****************************************/
        private:
            double phi( double x, int node, int der, int order );
            double quadR( double (hmFem::*f) (double), double a, double b, double eps, int maxIt);
            vec1d GLS( vec2d A, vec1d b);
            vec1d chol( vec2d A, vec1d b);
            vec1d gaussE( vec2d A, vec1d b);
            double det( vec2d A);
            vec2d inv(vec2d A);
            vec1d eig(vec2d A);
            bool isQuadr(vec2d A);
            bool isSymm(vec2d A);
            void lu_decomp(const vec2d &A, vec2d &L, vec2d &U);

            vec2d addToGSM( vec2d A_l, int e, vec2d A );
            vec1d addToGSV( vec1d v_l, int e, vec1d v );
            vec2d getK_loc( double (hmFem::*LHSvariationsF)(double x)  );
            vec2d getM_loc( );
            vec1d getF_loc( );
            double RHS_f(double x);
            double RHS_eig(double x);
            double LHS_sturmLiouville(double x);






    };



#endif  /* HMFEM_Included */
